@@ -14,11 +14,16 @@ from git_machete.exceptions import MacheteException
 from git_machete.utils import (AnsiEscapeCodes, CommandResult, colored, debug,
                                fmt)
 
+GITHUB_NEW_ISSUE_MESSAGE = 'Consider posting an issue at https://github.com/VirtusLab/git-machete/issues/new'
+
 
 class AnyRevision(str):
     @staticmethod
-    def of(value: str) -> Optional["AnyRevision"]:
-        return AnyRevision(value) if value else None
+    def of(value: str) -> "AnyRevision":
+        if not value:
+            raise TypeError(f'AnyRevision.of should not accept {value} as a param.\n' + GITHUB_NEW_ISSUE_MESSAGE)
+
+        return AnyRevision(value)
 
     def full_name(self) -> "AnyRevision":
         return self
@@ -26,8 +31,10 @@ class AnyRevision(str):
 
 class AnyBranchName(AnyRevision):
     @staticmethod
-    def of(value: str) -> Optional["AnyBranchName"]:
-        return AnyBranchName(value) if value else None
+    def of(value: str) -> "AnyBranchName":
+        if not value:
+            raise TypeError(f'AnyBranchName.of should not accept {value} as a param.\n' + GITHUB_NEW_ISSUE_MESSAGE)
+        return AnyBranchName(value)
 
     def full_name(self) -> "AnyBranchName":
         return self
@@ -35,33 +42,29 @@ class AnyBranchName(AnyRevision):
 
 class LocalBranchShortName(AnyBranchName):
     @staticmethod
-    def of(value: str) -> Optional["LocalBranchShortName"]:
-        if value:
-            if value.startswith('refs/heads/') or value.startswith('refs/remotes/'):
-                raise TypeError(
-                    f'LocalBranchShortName cannot accept refs/heads or refs/remotes. Given value: {value}.\n'
-                    'Consider posting an issue on https://github.com/VirtusLab/git-machete/issues/new')
-            else:
-                return LocalBranchShortName(value)
-        return None
+    def of(value: str) -> "LocalBranchShortName":
+        if value.startswith('refs/heads/') or value.startswith('refs/remotes/'):
+            raise TypeError(
+                f'LocalBranchShortName cannot accept `refs/heads` or `refs/remotes`. Given value: {value}.\n' + GITHUB_NEW_ISSUE_MESSAGE)
+        else:
+            return LocalBranchShortName(value)
 
-    def full_name(self) -> Optional["LocalBranchFullName"]:
+    def full_name(self) -> "LocalBranchFullName":
         return LocalBranchFullName.from_short_name(self)
 
 
 class LocalBranchFullName(AnyBranchName):
     @staticmethod
-    def of(value: str) -> Optional["LocalBranchFullName"]:
-        if value:
-            if value.startswith('refs/heads/'):
-                return LocalBranchFullName(value)
-            else:
-                raise TypeError(f'LocalBranchFullName needs to have `refs/heads` prefix before branch name. Given value: {value}.\n'
-                                'Consider posting an issue on https://github.com/VirtusLab/git-machete/issues/new')
-        return None
+    def of(value: str) -> "LocalBranchFullName":
+        if value and value.startswith('refs/heads/'):
+            return LocalBranchFullName(value)
+        else:
+            raise TypeError(
+                f'LocalBranchFullName needs to have `refs/heads` prefix before branch name. Given value: {value}.\n' +
+                GITHUB_NEW_ISSUE_MESSAGE)
 
     @staticmethod
-    def from_short_name(value: LocalBranchShortName) -> Optional["LocalBranchFullName"]:
+    def from_short_name(value: LocalBranchShortName) -> "LocalBranchFullName":
         return LocalBranchFullName.of(f"refs/heads/{value}")
 
     def full_name(self) -> "LocalBranchFullName":
@@ -73,37 +76,33 @@ class LocalBranchFullName(AnyBranchName):
 
 class RemoteBranchShortName(AnyBranchName):
     @staticmethod
-    def of(value: str) -> Optional["RemoteBranchShortName"]:
-        if value:
-            if value.startswith('refs/heads/') or value.startswith('refs/remotes/'):
-                raise TypeError(
-                    f'RemoteBranchShortName cannot accept refs/heads or refs/remotes. Given value: {value}.\n'
-                    'Consider posting an issue on https://github.com/VirtusLab/git-machete/issues/new')
-            else:
-                return RemoteBranchShortName(value)
-        return None
+    def of(value: str) -> "RemoteBranchShortName":
+        if value and value.startswith('refs/heads/') or value.startswith('refs/remotes/'):
+            raise TypeError(
+                f'RemoteBranchShortName cannot accept `refs/heads` or `refs/remotes`. Given value: {value}.\n' + GITHUB_NEW_ISSUE_MESSAGE)
+        else:
+            return RemoteBranchShortName(value)
 
-    def full_name(self) -> Optional["RemoteBranchFullName"]:
+    def full_name(self) -> "RemoteBranchFullName":
         return RemoteBranchFullName.from_short_name(self)
 
 
 class RemoteBranchFullName(AnyBranchName):
     @staticmethod
-    def of(value: str) -> Optional["RemoteBranchFullName"]:
-        if value:
-            if value.startswith('refs/remotes/'):
-                return RemoteBranchFullName(value)
-            else:
-                raise TypeError(f'RemoteBranchFullName needs to have `refs/remotes` prefix before branch name. Given value: {value}.\n'
-                                'Consider posting an issue on https://github.com/VirtusLab/git-machete/issues/new')
-        return None
+    def of(value: str) -> "RemoteBranchFullName":
+        if value and value.startswith('refs/remotes/'):
+            return RemoteBranchFullName(value)
+        else:
+            raise TypeError(
+                f'RemoteBranchFullName needs to have `refs/remotes` prefix before branch name. Given value: {value}.\n' +
+                GITHUB_NEW_ISSUE_MESSAGE)
 
     @staticmethod
     def is_valid(value: str) -> bool:
         return value is not None and value.startswith('refs/remotes/')
 
     @staticmethod
-    def from_short_name(value: RemoteBranchShortName) -> Optional["RemoteBranchFullName"]:
+    def from_short_name(value: RemoteBranchShortName) -> "RemoteBranchFullName":
         return RemoteBranchFullName.of(f"refs/remotes/{value}")
 
     def full_name(self) -> "RemoteBranchFullName":
@@ -115,19 +114,16 @@ class RemoteBranchFullName(AnyBranchName):
 
 class FullCommitHash(AnyRevision):
     @staticmethod
-    def of(value: str) -> Optional["FullCommitHash"]:
-        value = value.strip()
-        if value:
-            if len(value) == 40:
-                return FullCommitHash(value)
-            else:
-                raise TypeError(f'FullCommitHash requires length of 40. Given value: {value}, has length: {len(value)}.\n'
-                                'Consider posting an issue on https://github.com/VirtusLab/git-machete/issues/new')
-        return None
+    def of(value: str) -> "FullCommitHash":
+        if value and len(value) == 40:
+            return FullCommitHash(value)
+        else:
+            raise TypeError(
+                f'FullCommitHash requires length of 40. Given value: "{value}".\n' + GITHUB_NEW_ISSUE_MESSAGE)
 
     @staticmethod
-    def is_valid(value: str, git_context: "GitContext") -> bool:
-        return value is not None and len(value) == 40 and all(c in string.hexdigits for c in value)
+    def is_valid(value: str) -> bool:
+        return len(value) == 40 and all(c in string.hexdigits for c in value)
 
     def full_name(self) -> "FullCommitHash":
         return self
@@ -135,14 +131,13 @@ class FullCommitHash(AnyRevision):
 
 class ShortCommitHash(AnyRevision):
     @staticmethod
-    def of(value: str) -> Optional["ShortCommitHash"]:
-        if value:
-            if len(value) >= 7:
-                return ShortCommitHash(value)
-            else:
-                raise TypeError(f'ShortCommitHash requires length greater or equal to 7. Given value: {value}, has length: {len(value)}.\n'
-                                'Consider posting an issue on https://github.com/VirtusLab/git-machete/issues/new')
-        return None
+    def of(value: str) -> "ShortCommitHash":
+        if value and len(value) >= 7:
+            return ShortCommitHash(value)
+        else:
+            raise TypeError(
+                f'ShortCommitHash requires length greater or equal to 7. Given value: "{value}".\n' +
+                GITHUB_NEW_ISSUE_MESSAGE)
 
     def full_name(self) -> "ShortCommitHash":
         return self
@@ -151,7 +146,9 @@ class ShortCommitHash(AnyRevision):
 class FullTreeHash(str):
     @staticmethod
     def of(value: str) -> Optional["FullTreeHash"]:
-        return FullTreeHash(value) if value else None
+        if not value:
+            raise TypeError(f'FullTreeHash.of should not accept {value} as a param.\n' + GITHUB_NEW_ISSUE_MESSAGE)
+        return FullTreeHash(value)
 
 
 class ForkPointOverrideData:
@@ -200,8 +197,8 @@ class GitContext:
         self.__remote_branches_cached: Optional[List[RemoteBranchShortName]] = None
         self.__initial_log_hashes_cached: Dict[FullCommitHash, List[FullCommitHash]] = {}
         self.__remaining_log_hashes_cached: Dict[FullCommitHash, List[FullCommitHash]] = {}
-        self.__reflogs_cached: Optional[Dict[AnyBranchName, Optional[List[GitReflogEntry]]]] = None
-        self.__merge_base_cached: Dict[Tuple[FullCommitHash, FullCommitHash], FullCommitHash] = {}
+        self.__reflogs_cached: Optional[Dict[AnyBranchName, List[GitReflogEntry]]] = None
+        self.__merge_base_cached: Dict[Tuple[FullCommitHash, FullCommitHash], Optional[FullCommitHash]] = {}
         self.__is_equivalent_tree_reachable_cached: Dict[Tuple[FullCommitHash, FullCommitHash], bool] = {}
 
     @staticmethod
@@ -279,8 +276,10 @@ class GitContext:
             # We need to cut out the x.y.z part and not just take the result of 'git version' as is,
             # because the version string in certain distributions of git (esp. on OS X) has an extra suffix,
             # which is irrelevant for our purpose (checking whether certain git CLI features are available/bugs are fixed).
-            raw = re.search(r"\d+.\d+.\d+", self._popen_git("version").stdout).group(0)
-            self._git_version = tuple(map(int, raw.split(".")))
+            raw = re.search(r"\d+.\d+.\d+", self._popen_git("version").stdout)
+            if not raw:  # unlikely, never observed so far; mostly to satisfy mypy
+                return 0, 0, 0
+            self._git_version = tuple(map(int, raw.group(0).split(".")))
         return self._git_version
 
     def get_root_dir(self) -> str:
@@ -342,15 +341,16 @@ class GitContext:
 
     def get_config_attr_or_none(self, key: str) -> Optional[str]:
         self.__ensure_config_loaded()
+        assert self.__config_cached is not None
         return self.__config_cached.get(key.lower())
 
     def get_boolean_config_attr(self, key: str, default_value: bool) -> bool:
-        if self.get_boolean_config_attr_or_none(key) is not None:
-            return self.get_boolean_config_attr_or_none(key)
-        return default_value
+        value = self.get_boolean_config_attr_or_none(key)
+        return value if value is not None else default_value
 
     def get_boolean_config_attr_or_none(self, key: str) -> Optional[bool]:
         self.__ensure_config_loaded()
+        assert self.__config_cached is not None
         if self.__config_cached.get(key.lower()) is not None:
             return self.__config_cached.get(key.lower()) == 'true'
         return None
@@ -358,10 +358,12 @@ class GitContext:
     def set_config_attr(self, key: str, value: str) -> None:
         self._run_git("config", "--", key, value)
         self.__ensure_config_loaded()
+        assert self.__config_cached is not None
         self.__config_cached[key.lower()] = value
 
     def unset_config_attr(self, key: str) -> None:
         self.__ensure_config_loaded()
+        assert self.__config_cached is not None
         if self.get_config_attr_or_none(key):
             self._run_git("config", "--unset", key)
             del self.__config_cached[key.lower()]
@@ -375,8 +377,10 @@ class GitContext:
             self.__remotes_cached = utils.get_non_empty_lines(self._popen_git("remote").stdout)
         return self.__remotes_cached
 
-    def get_url_of_remote(self, remote: str) -> str:
-        return self.get_config_attr_or_none(f"remote.{remote}.url").strip()  # 'git remote get-url' method has only been added in git v2.5.1
+    def get_url_of_remote(self, remote: str) -> Optional[str]:
+        self.__ensure_config_loaded()
+        url = self.get_config_attr_or_none(f"remote.{remote}.url")  # 'git remote get-url' method has only been added in git v2.5.1
+        return url.strip() if url else None
 
     def fetch_remote(self, remote: str) -> None:
         if remote not in self.__fetch_done_for:
@@ -442,6 +446,7 @@ class GitContext:
             return FullCommitHash.of(revision)
         if self.__commit_hash_by_revision_cached is None:
             self.__load_branches()
+        assert self.__commit_hash_by_revision_cached is not None
         if revision not in self.__commit_hash_by_revision_cached:
             self.__commit_hash_by_revision_cached[revision] = self.__find_commit_hash_by_revision(revision)
         return self.__commit_hash_by_revision_cached[revision]
@@ -455,6 +460,7 @@ class GitContext:
     def get_tree_hash_by_commit_hash(self, commit_hash: FullCommitHash) -> Optional[FullTreeHash]:
         if self.__tree_hash_by_commit_hash_cached is None:
             self.__load_branches()
+        assert self.__tree_hash_by_commit_hash_cached is not None
         if commit_hash not in self.__tree_hash_by_commit_hash_cached:
             self.__tree_hash_by_commit_hash_cached[commit_hash] = self.__find_tree_hash_by_revision(commit_hash)
         return self.__tree_hash_by_commit_hash_cached[commit_hash]
@@ -469,6 +475,7 @@ class GitContext:
     def get_committer_unix_timestamp_by_revision(self, revision: AnyBranchName) -> int:
         if self.__committer_unix_timestamp_by_revision_cached is None:
             self.__load_branches()
+        assert self.__committer_unix_timestamp_by_revision_cached is not None
         return self.__committer_unix_timestamp_by_revision_cached.get(revision.full_name(), 0)
 
     def __get_remotes_containing_branch(self, branch: LocalBranchShortName, remotes: Optional[List[str]] = None) -> List[str]:
@@ -513,6 +520,7 @@ class GitContext:
     def get_strict_counterpart_for_fetching_of_branch(self, branch: LocalBranchShortName) -> Optional[RemoteBranchShortName]:
         if self.__counterparts_for_fetching_cached is None:
             self.__load_branches()
+        assert self.__counterparts_for_fetching_cached is not None
         return self.__counterparts_for_fetching_cached.get(branch)
 
     def get_combined_counterpart_for_fetching_of_branch(self, branch: LocalBranchShortName) -> Optional[RemoteBranchShortName]:
@@ -543,11 +551,13 @@ class GitContext:
     def get_local_branches(self) -> List[LocalBranchShortName]:
         if self.__local_branches_cached is None:
             self.__load_branches()
+        assert self.__local_branches_cached is not None
         return self.__local_branches_cached
 
     def get_remote_branches(self) -> List[RemoteBranchShortName]:
         if self.__remote_branches_cached is None:
             self.__load_branches()
+        assert self.__remote_branches_cached is not None
         return self.__remote_branches_cached
 
     def __load_branches(self) -> None:
@@ -621,10 +631,16 @@ class GitContext:
         # %gd - reflog selector (refname@{num})
         # %H - full hash
         # %gs - reflog subject
-        local_branches = [str(branch.full_name()) for branch in self.get_local_branches()]  # str here to match _popen_git() input type
-        counterpart_branches = [str(self.get_combined_counterpart_for_fetching_of_branch(branch).full_name()) for branch in
-                                self.get_local_branches() if self.get_combined_counterpart_for_fetching_of_branch(branch)]
-        all_branches = local_branches + counterpart_branches
+        local_branches: List[str] = [branch.full_name() for branch in self.get_local_branches()]  # str to match _popen_git() input type
+
+        def get_counterpart_branches() -> Generator[str, None, None]:
+            for branch in self.get_local_branches():
+                counterpart = self.get_combined_counterpart_for_fetching_of_branch(branch)
+                if counterpart:
+                    yield counterpart.full_name()
+        counterpart_branches: List[str] = list(get_counterpart_branches())
+
+        all_branches: List[str] = local_branches + counterpart_branches
 
         # The trailing '--' is necessary to avoid ambiguity in case there is a file called just exactly like one of the branches.
         entries = utils.get_non_empty_lines(self._popen_git("reflog", "show", "--format=%gD\t%H\t%gs", *(all_branches + ["--"])).stdout)
@@ -638,9 +654,10 @@ class GitContext:
             if len(branch_and_pos) != 2:  # invalid, shouldn't happen
                 continue
             branch, pos = branch_and_pos
-            if branch not in self.__reflogs_cached:
-                self.__reflogs_cached[AnyBranchName.of(branch)] = []
-            self.__reflogs_cached[AnyBranchName.of(branch)] += [GitReflogEntry(hash=FullCommitHash.of(hash), reflog_subject=subject)]
+            any_branch_name = AnyBranchName.of(branch)
+            if any_branch_name not in self.__reflogs_cached:
+                self.__reflogs_cached[any_branch_name] = []
+            self.__reflogs_cached[any_branch_name] += [GitReflogEntry(hash=FullCommitHash.of(hash), reflog_subject=subject)]
 
     def get_reflog(self, branch: AnyBranchName) -> List[GitReflogEntry]:
         # git version 2.14.2 fixed a bug that caused fetching reflog of more than
@@ -648,6 +665,7 @@ class GitContext:
         if self.get_git_version() >= (2, 14, 2):
             if self.__reflogs_cached is None:
                 self.__load_all_reflogs()
+            assert self.__reflogs_cached is not None
             return self.__reflogs_cached.get(branch, [])
         else:
             if self.__reflogs_cached is None:
@@ -749,7 +767,7 @@ class GitContext:
             raise MacheteException("Not currently on any branch")
         return result
 
-    def __get_merge_base_for_commit_hashes(self, hash1: FullCommitHash, hash2: FullCommitHash) -> FullCommitHash:
+    def __get_merge_base_for_commit_hashes(self, hash1: FullCommitHash, hash2: FullCommitHash) -> Optional[FullCommitHash]:
         # This if statement is not changing the outcome of the later return, but
         # it enhances the efficiency of the script. If both hashes are the same,
         # there is no point running git merge-base.
@@ -768,8 +786,8 @@ class GitContext:
             # In the rare case when hash1, hash2 have no common commits, the flag: allow_non_zero=True
             # (allows, non zero exit code to be returned by git merge-base command, without raising an exception)
             # is used and the __get_merge_base function returns None.
-            self.__merge_base_cached[hash1, hash2] = FullCommitHash.of(self._popen_git("merge-base", hash1, hash2,
-                                                                                       allow_non_zero=True).stdout)
+            merge_base = self._popen_git("merge-base", hash1, hash2, allow_non_zero=True).stdout.strip()
+            self.__merge_base_cached[hash1, hash2] = FullCommitHash.of(merge_base) if merge_base else None
         return self.__merge_base_cached[hash1, hash2]
 
     # Note: the 'git rev-parse --verify' validation is not performed in case for either of earlier/later
@@ -781,6 +799,8 @@ class GitContext:
     ) -> bool:
         earlier_hash = self.get_commit_hash_by_revision(earlier_revision)
         later_hash = self.get_commit_hash_by_revision(later_revision)
+        if not earlier_hash or not later_hash:
+            return False
         return self.__get_merge_base_for_commit_hashes(earlier_hash, later_hash) == earlier_hash
 
     def get_merge_base(
@@ -790,6 +810,8 @@ class GitContext:
     ) -> Optional[FullCommitHash]:
         earlier_hash = self.get_commit_hash_by_revision(earlier_revision)
         later_hash = self.get_commit_hash_by_revision(later_revision)
+        if not earlier_hash or not later_hash:
+            return None
         return self.__get_merge_base_for_commit_hashes(earlier_hash, later_hash)
 
     # Determine if reachable_from, or any ancestors of reachable_from that are NOT ancestors of equivalent_to,
@@ -802,6 +824,8 @@ class GitContext:
     ) -> bool:
         equivalent_to_commit_hash = self.get_commit_hash_by_revision(equivalent_to)
         reachable_from_commit_hash = self.get_commit_hash_by_revision(reachable_from)
+        if not equivalent_to_commit_hash or not reachable_from_commit_hash:
+            return False
 
         if equivalent_to_commit_hash == reachable_from_commit_hash:
             return True
@@ -886,7 +910,7 @@ class GitContext:
         self._run_git("merge", "--ff-only", branch.full_name())
         self.flush_caches()
 
-    def rebase(self, onto: AnyRevision, fork_revision: AnyRevision, branch: LocalBranchShortName, opt_no_interactive_rebase: bool) -> None:
+    def rebase(self, onto: AnyRevision, from_exclusive: AnyRevision, branch: LocalBranchShortName, opt_no_interactive_rebase: bool) -> None:
         def do_rebase() -> None:
             # Let's use `OPTS` suffix for consistency with git's built-in env var `GIT_DIFF_OPTS`
             git_machete_rebase_opts_var = 'GIT_MACHETE_REBASE_OPTS'
@@ -894,7 +918,7 @@ class GitContext:
             try:
                 if not opt_no_interactive_rebase:
                     rebase_opts.append("--interactive")
-                self._run_git("rebase", *rebase_opts, "--onto", onto, fork_revision, branch)
+                self._run_git("rebase", *rebase_opts, "--onto", onto, from_exclusive, branch)
             finally:
                 # https://public-inbox.org/git/317468c6-40cc-9f26-8ee3-3392c3908efb@talktalk.net/T
                 # In our case, this can happen when git version invoked by git-machete to start the rebase
@@ -926,7 +950,7 @@ class GitContext:
         hook_path = self.get_hook_path("machete-pre-rebase")
         if self.check_hook_executable(hook_path):
             debug(f"running machete-pre-rebase hook ({hook_path})")
-            exit_code = utils.run_cmd(hook_path, onto, fork_revision, branch, cwd=self.get_root_dir())
+            exit_code = utils.run_cmd(hook_path, onto, from_exclusive, branch, cwd=self.get_root_dir())
             if exit_code == 0:
                 do_rebase()
             else:
@@ -1011,7 +1035,7 @@ class GitContext:
         params = ["log", "-1", f"--format={information.value}"]
         if commit:
             params.append(commit)
-        return self._popen_git(*params).stdout
+        return self._popen_git(*params).stdout.strip()
 
     def display_branch_history_from_fork_point(self, branch: LocalBranchFullName, fork_point: FullCommitHash) -> int:
         return self._run_git("log", f"^{fork_point}", branch)
@@ -1020,14 +1044,14 @@ class GitContext:
             self, parent_revision: AnyRevision, msg: str, env: Dict[str, str]) -> FullCommitHash:
         # returns hash of the new commit
         return FullCommitHash.of(self._popen_git(
-            "commit-tree", "HEAD^{tree}", "-p", parent_revision, "-m", msg, env=env).stdout)  # noqa: FS003
+            "commit-tree", "HEAD^{tree}", "-p", parent_revision, "-m", msg, env=env).stdout.strip())  # noqa: FS003
 
     def delete_branch(self, branch_name: LocalBranchShortName, force: bool = False) -> int:
         self.flush_caches()
         delete_option = '-D' if force else '-d'
         return self._run_git("branch", delete_option, branch_name)
 
-    def display_diff(self, fork_point: AnyRevision, format_with_stat: bool, branch: LocalBranchShortName = None) -> int:
+    def display_diff(self, fork_point: AnyRevision, format_with_stat: bool, branch: Optional[LocalBranchShortName] = None) -> int:
         params = ["diff"]
         if format_with_stat:
             params.append("--stat")
